""" Freeze variables and convert 2 generator networks to 2 GraphDef files.
This makes file size smaller and can be used for inference in production.
An example of command-line usage is:
python export_graph.py --checkpoint_dir checkpoints/20170424-1152 \
                       --XtoY_model apple2orange.pb \
                       --YtoX_model orange2apple.pb \
                       --image_size 256
"""

import tensorflow as tf
import os

from model import GAN
from utils import image_utils

FLAGS = tf.flags.FLAGS

tf.flags.DEFINE_string('checkpoint_dir', './checkpoints/20190709-2146_semi-5000_G-lr-0.0001_D-lr-0.0001_yas-0.01_yau-0.0001_yu-0.001_t-0.2', 'checkpoints directory path')
tf.flags.DEFINE_string('model', 'model.pb', 'model name')

# Utilized data
tf.flags.DEFINE_string('training_file', '../../data/tfrecords/VOC2012/train_n_divided_2.tfrecords',
                       'tfrecords file for training')
tf.flags.DEFINE_string('validation_file', '../../data/tfrecords/VOC2012/val.tfrecords', 'tfrecords file for validation')
tf.flags.DEFINE_string('unsupervised_file', '../../data/tfrecords/VOC2012/unsupervised_n_divided_2_label.tfrecords',
                       'tfrecords file for unsupervised learning')
tf.flags.DEFINE_string('pretrained_folder', '../../data/pretrained', 'path of the folder for pretraied encoder models')

# Data format
tf.flags.DEFINE_integer('batch_size', 1, 'batch size')
tf.flags.DEFINE_integer('validation_batch_size', 1, ' validation batch size default: 1449')
tf.flags.DEFINE_integer('image_height', 375, 'image size, default: 256')
tf.flags.DEFINE_integer('image_width', 750, 'image size, default: 256')
tf.flags.DEFINE_integer('num_classes', 19, 'number of output classes')
tf.flags.DEFINE_string('norm', 'batch', '[instance, batch] use instance norm or batch norm, default: instance')

# Training parameters
tf.flags.DEFINE_string('train_type', 'adversarial',
                       '[supervised, adversarial, semi-supervised, after-supervised] type of training to use')
tf.flags.DEFINE_integer('training_steps', 20006, 'number of training steps')
tf.flags.DEFINE_integer('logging_steps_interval', 1, 'interval of steps between each log')
tf.flags.DEFINE_integer('validation_steps_interval', 1, 'interval of steps between each validation')
tf.flags.DEFINE_integer('save_steps_interval', 10000, 'interval of steps between each log')
tf.flags.DEFINE_integer('supervised_steps', 3, 'number of supervised steps before starting unsupervided training')
tf.flags.DEFINE_boolean('use_lsgan', False,
                        'define the type of loss of the discriminator (least square or cross_entropy)')

# Checkpoint folder
tf.flags.DEFINE_string('load_model', '20181203-1525',
                       'folder of saved model that you wish to continue training (checkpoint/{})')

# Network optimization parameters
tf.flags.DEFINE_float('lambda_adv_sup', 0.01, 'weight for discriminator in supervised mode')
tf.flags.DEFINE_float('lambda_adv_unsup', 0.001, 'weight for discriminator in unsupervised mode')
tf.flags.DEFINE_float('lambda_unsup', 0.1, 'weight for discriminator')
tf.flags.DEFINE_float('threshold_unsup', 0.2, 'threshold applied to confidence map produced by discriminator')

# Generator optimization parameter (generator uses SGD optimizer)
tf.flags.DEFINE_float('generator_learning_rate', 2.5e-4, 'initial learning rate for generator')
tf.flags.DEFINE_float('generator_end_learning_rate', 0.0, 'final decay rate for generator optimizer')
tf.flags.DEFINE_integer('generator_start_decay_step', 0, 'number of steps after which start decay for generator')
tf.flags.DEFINE_integer('generator_decay_steps', 20000, 'number of steps of decay learning rate for generator')
tf.flags.DEFINE_float('generator_decay_power', 0.9, 'power of weight decay for generator optimizer')
tf.flags.DEFINE_float('generator_momentum', 0.9, 'momentum for generator optimizer')
tf.flags.DEFINE_float('generator_weight_decay', 1e-4, 'weight decay for momentum of generator optimizer')
tf.flags.DEFINE_boolean('save_summary_images', False,
                        'specifies if the training images must be saved during the training (NB: higher memory usage)')

# Discriminator optimization parameter (discriminator uses Adam optimizer)
tf.flags.DEFINE_float('discriminator_learning_rate', 1e-4, 'initial learning rate for discriminator')
tf.flags.DEFINE_float('discriminator_end_learning_rate', 0.0, 'final decay rate for discriminator optimizer')
tf.flags.DEFINE_integer('discriminator_start_decay_step', 0,
                        'number of steps after which start decay for discriminator')
tf.flags.DEFINE_integer('discriminator_decay_steps', 20000, 'number of steps of decay learning rate for discriminator')
tf.flags.DEFINE_float('discriminator_decay_power', 0.9, 'power of weight decay for discriminator optimizer')
tf.flags.DEFINE_float('discriminator_momentum', 0.9, 'momentum for generator optimizer (utilized as beta1 parameter)')
tf.flags.DEFINE_float('discriminator_weight_decay', 1e-4, 'weight decay for discriminator optimizer')

def export_graph(model_name):
    try:
        os.makedirs('output_model')
    except os.error:
        pass

    graph = tf.Graph()

    # define mapping vector for domain adaptation
    #             [0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20]
    # class_mask = [0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0]
    class_mask = None

    class_weights_Cityscapes_Val = [0.629240815350735, 0.940728658835580, 0.765487700112771, 0.992872888348501,
                                    0.990494855410113, 0.988287167043793, 0.998120863476552, 0.994759080991432,
                                    0.845413256533045, 0.989703596027120, 0.960282326319566, 0.988891268487100,
                                    0.998375382630173, 0.931349734636790, 0.996732388535060, 0.998320191040540,
                                    0.996800181800288, 0.998813054898909, 0.995326589521930]

    class_weights_GTA_Val = [0.637960276015760, 0.906876401511850, 0.808888109204147, 0.979417971165853,
                             0.992849926507391, 0.988065518323295, 0.998489616496333, 0.999064071693913,
                             0.914391166506145, 0.975514686906170, 0.848631300794846, 0.995948128310475,
                             0.999638046230835, 0.971915722873151, 0.987760747283702, 0.995798327177682,
                             0.999222562735153, 0.999635347188153, 0.999932073075146]

    with graph.as_default():
        tf.set_random_seed(12345)
        gan = GAN(
            train_file=FLAGS.training_file,
            validation_file=FLAGS.validation_file,
            unsupervised_file=FLAGS.unsupervised_file,
            batch_size=FLAGS.batch_size,
            image_height=FLAGS.image_height,
            image_width=FLAGS.image_width,
            num_classes=FLAGS.num_classes,
            pretrained_folder=FLAGS.pretrained_folder,
            lambda_adv_sup=FLAGS.lambda_adv_sup,
            lambda_adv_unsup=FLAGS.lambda_adv_unsup,
            lambda_unsup=FLAGS.lambda_unsup,
            threshold_unsup=FLAGS.threshold_unsup,
            generator_learning_rate=FLAGS.generator_learning_rate,
            generator_momentum=FLAGS.generator_momentum,
            generator_weight_decay=FLAGS.generator_weight_decay,
            generator_decay_power=FLAGS.generator_decay_power,
            generator_decay_steps=FLAGS.generator_decay_steps,
            generator_start_decay_step=FLAGS.generator_start_decay_step,
            generator_end_learning_rate=FLAGS.generator_end_learning_rate,
            discriminator_learning_rate=FLAGS.discriminator_learning_rate,
            discriminator_momentum=FLAGS.discriminator_momentum,
            discriminator_weight_decay=FLAGS.discriminator_weight_decay,
            discriminator_decay_power=FLAGS.discriminator_decay_power,
            discriminator_decay_steps=FLAGS.discriminator_decay_steps,
            discriminator_start_decay_step=FLAGS.discriminator_start_decay_step,
            discriminator_end_learning_rate=FLAGS.discriminator_end_learning_rate,
            train_type=FLAGS.train_type,
            use_lsgan=FLAGS.use_lsgan,
            save_summary_images=FLAGS.save_summary_images,
            class_weights=class_weights_GTA_Val
        )

        image_data = tf.placeholder(tf.string, name='input_image')
        input_image = tf.image.decode_jpeg(image_data, channels=3)
        input_image = tf.image.resize_images(input_image, size=(FLAGS.image_height, FLAGS.image_width))
        input_image = image_utils.convert2float(input_image)
        input_image.set_shape([FLAGS.image_height, FLAGS.image_width, 3])
        output_image = gan.G.sample(tf.expand_dims(input_image, 0))

        output_image = tf.identity(output_image, name='output_image')
        restore_saver = tf.train.Saver()

    with tf.Session(graph=graph) as sess:
        sess.run(tf.global_variables_initializer())
        latest_ckpt = tf.train.latest_checkpoint(FLAGS.checkpoint_dir)
        restore_saver.restore(sess, latest_ckpt)
        output_graph_def = tf.graph_util.convert_variables_to_constants(
            sess, graph.as_graph_def(), [output_image.op.name])

        tf.train.write_graph(output_graph_def, 'output_model', model_name, as_text=False)

def main(unused_argv):
    print('Export model...')
    export_graph(FLAGS.model)


if __name__ == '__main__':
    tf.app.run()
