import tensorflow as tf
from tensorflow.python.ops.image_ops_impl import ResizeMethod
from tensorflow.python.ops.losses.losses_impl import Reduction

from discriminator import Discriminator
from generator import Generator
from reader_supervised import ReaderSupervised
from reader_unsupervised import ReaderUnsupervised
from utils import image_utils, tensor_utils

REAL_LABEL = 0.9


class GAN:
    def __init__(self,
                 train_file,
                 validation_file,
                 unsupervised_file,
                 batch_size,
                 image_height,
                 image_width,
                 num_classes,
                 pretrained_folder,
                 lambda_adv_sup,
                 lambda_adv_unsup,
                 lambda_unsup,
                 threshold_unsup,
                 generator_learning_rate,
                 generator_momentum,
                 generator_weight_decay,
                 generator_decay_power,
                 generator_decay_steps,
                 generator_start_decay_step,
                 generator_end_learning_rate,
                 discriminator_learning_rate,
                 discriminator_momentum,
                 discriminator_weight_decay,
                 discriminator_decay_power,
                 discriminator_decay_steps,
                 discriminator_start_decay_step,
                 discriminator_end_learning_rate,
                 train_type,
                 use_lsgan,
                 save_summary_images,
                 class_weights
                 ):

        self.train_file = train_file
        self.validation_file = validation_file
        self.unsupervised_file = unsupervised_file
        self.batch_size = batch_size
        self.image_height = image_height
        self.image_width = image_width
        self.num_classes = num_classes
        self.pretrained_folder = pretrained_folder
        self.lambda_adv_sup = lambda_adv_sup
        self.lambda_adv_unsup = lambda_adv_unsup
        self.lambda_unsup = lambda_unsup
        self.threshold_unsup = threshold_unsup
        self.generator_learning_rate = generator_learning_rate
        self.generator_momentum = generator_momentum
        self.generator_weight_decay = generator_weight_decay
        self.generator_decay_power = generator_decay_power
        self.generator_decay_steps = generator_decay_steps
        self.generator_start_decay_step = generator_start_decay_step
        self.generator_end_learning_rate = generator_end_learning_rate
        self.discriminator_learning_rate = discriminator_learning_rate
        self.discriminator_momentum = discriminator_momentum
        self.discriminator_weight_decay = discriminator_weight_decay
        self.discriminator_decay_power = discriminator_decay_power
        self.discriminator_decay_steps = discriminator_decay_steps
        self.discriminator_start_decay_step = discriminator_start_decay_step
        self.discriminator_end_learning_rate = discriminator_end_learning_rate
        self.train_type = train_type
        self.use_lsgan = use_lsgan
        self.save_summary_images = save_summary_images
        self.class_weights = tf.convert_to_tensor(class_weights, dtype=tf.float32)

        # Build network structure
        self.G = Generator(name='G', image_height=self.image_height, image_width=self.image_width,
                           num_classes=num_classes,
                           pretrained_folder=pretrained_folder, weight_decay=self.generator_weight_decay)
        if train_type == 'adversarial' or train_type == 'semi-supervised' or train_type == 'domain-adaptation':
            self.D = Discriminator(name='D', image_height=self.image_height, image_width=self.image_width)

        self.lambda_semi = tf.placeholder(dtype=tf.float32, shape=[], name='temp')

        self.unsup_input = tf.placeholder(tf.float32, [batch_size, image_height, image_width, 3])
        self.expanded_mask = tf.placeholder(tf.float32, [batch_size, image_height, image_width])
    def training_model(self):
        """
        Build the tensorflow graph composed by generator and discriminator

        :return: parameters needed to perform the training
        """
        # Initialize variables
        G_sup_loss = 0
        G_unsup_loss_tot = 0
        G_unsup_loss_adv = 0
        G_unsup_loss_ce = 0
        D_loss = 0
        D_loss_fake = 0
        D_loss_real = 0

        training_summary_sup_list = []
        training_summary_unsup_list = []
        D_summary = None
        G_sup_loss_summary = None
        G_unsup_loss_summary = None
        training_summary_unsup = None
        G_optimizer_sup = None
        G_unsup_adv_optimizer = None
        G_unsup_tot_optimizer = None
        D_optimizer = None
        G_optim_summary = None
        G_unsup_adv_optim_summary = None
        G_unsup_adv_semi_optim_summary = None
        D_optim_summary = None
        G_composed_loss_adv = None
        G_composed_loss_adv_semi = None
        G_optimizer_unsup_adv = None
        G_optimizer_unsup_adv_semi = None
        G_output_unsup = None
        threshold_mask = None
        G_in_unsup = None
        G_output_unsup_softmax = None

        # Define readers
        reader_supervised_G = ReaderSupervised(self.train_file, name='G_sup', image_height=self.image_height,
                                               image_width=self.image_width, batch_size=self.batch_size,
                                               min_queue_examples=1000, seed=12345,
                                               save_summary_images=self.save_summary_images)
        reader_supervised_D = ReaderSupervised(self.train_file, name='D_sup',
                                               image_height=self.image_height,
                                               image_width=self.image_width,
                                               batch_size=self.batch_size, min_queue_examples=1000,
                                               seed=54321, save_summary_images=self.save_summary_images)
        reader_unsupervised = ReaderUnsupervised(self.unsupervised_file, name='unsup', image_height=self.image_height,
                                                 image_width=self.image_width, batch_size=self.batch_size,
                                                 min_queue_examples=1000, save_summary_images=self.save_summary_images)

        # Read input and output data from supervised dataset
        G_in_sup, G_labels, G_reader_summary = reader_supervised_G.feed()
        D_in_sup, D_labels, _ = reader_supervised_D.feed()

        # Define loss functions in different cases
        if self.train_type == 'supervised':
            G_sup_loss, G_sup_loss_summary = self.generator_loss_supervised(G_input=G_in_sup, gt_labels=G_labels)

        elif self.train_type == 'adversarial':
            G_sup_loss, G_sup_loss_summary = self.generator_loss_adv(G_input=G_in_sup, gt_labels=G_labels,
                                                                     lambda_adv=self.lambda_adv_sup)
            D_loss, D_summary = self.discriminator_loss(G_input_sup=D_in_sup, gt_labels=D_labels)

        elif self.train_type == 'semi-supervised' or self.train_type == 'domain-adaptation':
            G_in_unsup, unsup_reader_summary = reader_unsupervised.feed()
            G_sup_loss, G_sup_loss_summary = self.generator_loss_adv(G_input=G_in_sup, gt_labels=G_labels,
                                                                     lambda_adv=self.lambda_adv_sup)
            D_loss, D_summary = self.discriminator_loss(G_input_sup=D_in_sup, gt_labels=D_labels,
                                                        G_input_unsup=G_in_unsup)

            G_unsup_loss_adv, G_unsup_loss_ce, G_unsup_loss_summary = self.generator_loss_unsupervised()

            # G_composed_loss_adv = G_sup_loss + G_unsup_loss_adv
            G_composed_loss_adv_semi = G_sup_loss + G_unsup_loss_adv + G_unsup_loss_ce

            G_output_unsup = self.G(G_in_unsup)
            D_out_fake = self.D(tf.nn.softmax(G_output_unsup))
            D_out_sigmoid = tf.nn.sigmoid(D_out_fake)
            threshold_mask = tf.squeeze(tf.greater_equal(D_out_sigmoid, self.threshold_unsup), axis=3)
            G_output_unsup_softmax = tf.nn.softmax(G_output_unsup)
            G_output_unsup_softmax_RGB = image_utils.convert_output2rgb(G_output_unsup_softmax)

            if self.save_summary_images:
                training_summary_unsup_list.append(unsup_reader_summary)

        # Define optimizers in different cases
        if self.train_type == 'supervised':
            G_optimizer_sup, G_optim_summary = self.generator_optimizer(name='G_optimizer_sup', loss=G_sup_loss)

        elif self.train_type == 'adversarial':
            G_optimizer_sup, G_optim_summary = self.generator_optimizer(name='G_optimizer_sup', loss=G_sup_loss)
            D_optimizer, D_optim_summary = self.discriminator_optimizer(name='D_optimizer', loss=D_loss)

        elif self.train_type == 'semi-supervised' or self.train_type == 'domain-adaptation':
            G_optimizer_unsup_adv_semi, G_unsup_adv_semi_optim_summary = self.generator_optimizer(
                name='G_optimizer_adv_semi',
                loss=G_composed_loss_adv_semi)

            D_optimizer, D_optim_summary = self.discriminator_optimizer(name='D_optimizer', loss=D_loss)

        # Define summary in different cases
        if self.train_type == 'supervised':
            training_summary_sup_list.append(G_sup_loss_summary)
            training_summary_sup_list.append(G_optim_summary)

        elif self.train_type == 'adversarial':
            training_summary_sup_list.append(G_sup_loss_summary)
            training_summary_sup_list.append(G_optim_summary)
            training_summary_sup_list.append(D_summary)
            training_summary_sup_list.append(D_optim_summary)

        elif self.train_type == 'semi-supervised' or self.train_type == 'domain-adaptation':
            training_summary_sup_list.append(G_sup_loss_summary)
            training_summary_sup_list.append(D_summary)
            training_summary_sup_list.append(D_optim_summary)
            training_summary_unsup_list.append(G_unsup_loss_summary)
            # training_summary_unsup_list.append(G_unsup_adv_optim_summary)
            training_summary_unsup_list.append(G_unsup_adv_semi_optim_summary)
            training_summary_unsup = tf.summary.merge(training_summary_unsup_list)

        if self.save_summary_images:
            training_summary_sup_list.append(G_reader_summary)
        training_summary_sup = tf.summary.merge(training_summary_sup_list)

        return G_sup_loss, G_composed_loss_adv, G_composed_loss_adv_semi, D_loss, G_optimizer_sup, G_optimizer_unsup_adv_semi, D_optimizer, training_summary_sup, training_summary_unsup, G_in_unsup, G_output_unsup_softmax, G_output_unsup_softmax_RGB, threshold_mask

    def validation_model(self):
        """
        Build the tensorflow graph used to compute prediction scores on validation data

        :return: parameters needed to perform the validation
        """

        def _extract_fn(tfrecord):
            """
            Utility function used to extract data for validation

            :param tfrecord: file containing validation data
            :return: single data contained in the tfrecord file
            """
            # Extract features using the keys set during creation
            features = {
                'height': tf.FixedLenFeature([], tf.int64),
                'width': tf.FixedLenFeature([], tf.int64),
                'input': tf.FixedLenFeature([], tf.string),
                'gt': tf.FixedLenFeature([], tf.string)
            }

            # Obtain back image and labels
            sample = tf.parse_single_example(tfrecord, features)
            height = tf.cast(sample['height'], tf.int32)
            width = tf.cast(sample['width'], tf.int32)
            image = tf.image.decode_jpeg(sample['input'], channels=3)
            gt = tf.image.decode_png(sample['gt'], channels=1)

            # Preprocess paper_images
            image, gt = image_utils.preprocess_val_data(image, gt)
            return [image, gt]

        # Build iterator for the validation dataset
        validation_dataset = tf.data.TFRecordDataset(self.validation_file).map(_extract_fn)
        iterator = validation_dataset.make_initializable_iterator()
        next_val_data = iterator.get_next()

        # Obtain validation data
        val_image = tf.expand_dims(next_val_data[0], axis=0)
        val_label = tf.expand_dims(next_val_data[1], axis=0)

        # Validation loss
        resized_input = tf.image.resize_images(val_image, [self.image_height, self.image_width],
                                               method=ResizeMethod.BILINEAR)
        G_output = self.G(resized_input)
        G_output = tf.image.resize_images(G_output, [1024, 2048], method=ResizeMethod.BILINEAR)
        gt_one_hot = tensor_utils.convert_label2onehot(val_label, self.num_classes)
        loss = tf.nn.softmax_cross_entropy_with_logits_v2(logits=G_output, labels=gt_one_hot)
        target_mask = tf.squeeze(tf.less_equal(val_label, self.num_classes - 1), axis=3)
        loss_masked = tf.boolean_mask(loss, target_mask)
        validation_loss = tf.reduce_mean(loss_masked)

        # Define mean function
        mean_validation_loss, mean_validation_update_op = tf.metrics.mean(validation_loss)

        # Mean IoU
        mean_IoU, mean_IoU_update_op, accuracy, accuracy_update_op, confusion_matrix = self.compute_evaluation_metrics(
            G_input=val_image, gt_labels=val_label)

        # Summary
        validation_summary_list = [tf.summary.scalar('validation/validation_loss', mean_validation_loss),
                                   tf.summary.scalar('validation/mean_IoU', mean_IoU),
                                   tf.summary.scalar('validation/pixel_accuracy', accuracy)]
        validation_summary = tf.summary.merge(validation_summary_list)

        return iterator, mean_validation_loss, mean_validation_update_op, confusion_matrix, accuracy, accuracy_update_op, mean_IoU, mean_IoU_update_op, validation_summary

    def discriminator_optimizer(self, name, loss):
        """
        Define an optimizer for the discriminator

        :param name: name of the optimizer
        :param loss: loss to optimize
        :return: optimizer
        """
        global_step = tf.Variable(0, trainable=False)
        learning_rate = (
            tf.where(
                tf.greater_equal(global_step, self.discriminator_start_decay_step),
                tf.train.polynomial_decay(self.discriminator_learning_rate,
                                          global_step - self.discriminator_start_decay_step,
                                          self.discriminator_decay_steps, self.discriminator_end_learning_rate,
                                          power=self.discriminator_decay_power),
                self.discriminator_learning_rate
            )
        )
        summary = tf.summary.scalar('learning_rate/Adam_{}'.format(name), learning_rate)

        learning_step = (
            tf.train.AdamOptimizer(learning_rate, beta1=self.discriminator_momentum, name=name)
                .minimize(loss, global_step=global_step, var_list=self.D.variables)
        )

        return learning_step, summary

    def generator_optimizer(self, name, loss):
        """
        Define an optimizer for the generator

        :param name: name of the optimizer
        :param loss: loss to optimize
        :return: optimizer
        """
        global_step = tf.Variable(0, trainable=False)

        learning_rate = (
            tf.where(
                tf.greater_equal(global_step, self.generator_start_decay_step),
                tf.train.polynomial_decay(self.generator_learning_rate, global_step - self.generator_start_decay_step,
                                          self.generator_decay_steps, self.generator_end_learning_rate,
                                          power=self.generator_decay_power),
                self.generator_learning_rate
            )
        )
        summary = tf.summary.scalar('learning_rate/SGD_{}'.format(name), learning_rate)

        learning_step = (
            tf.train.MomentumOptimizer(learning_rate, momentum=self.generator_momentum, name=name)
                .minimize(loss + self.G.l2_loss, global_step=global_step, var_list=self.G.all_var)
        )
        return learning_step, summary

    def generator_loss_supervised(self, G_input, gt_labels):
        """
        Compute the loss of the generator in a supervised manner

        :param G_input: input of the generator. 4D tensor: [batch_size, image_width, image_height, 3]
        :param gt_labels: gt with classes. 4D tensor: [batch_size, image_width, image_height, 1]
        :return: supervised loss of the generator
        """
        # Compute needed variables
        G_output = self.G(G_input)
        gt_one_hot = tensor_utils.convert_label2onehot(gt_labels, self.num_classes)

        # Compute the cross entropy loss
        loss = tf.nn.softmax_cross_entropy_with_logits_v2(logits=G_output, labels=gt_one_hot)
        target_mask = tf.squeeze(tf.less_equal(gt_labels, self.num_classes - 1), axis=3)
        loss_masked = tf.boolean_mask(loss, target_mask)
        loss = tf.reduce_mean(loss_masked)

        # Summary
        summary = []
        if self.save_summary_images:
            image_slice = image_utils.convert_output2rgb(
                tf.slice(G_output, [0, 0, 0, 0], [1, self.image_height, self.image_width, self.num_classes]))
            summary = [tf.summary.image('G_sup/output', image_slice)]

        summary.append(tf.summary.scalar('loss_G/G', loss))
        return loss, summary

    def discriminator_loss(self, G_input_sup, gt_labels, G_input_unsup=None):
        """
        Compute the loss of the discriminator

        :param G_input_sup: supervised input of the generator. 4D tensor: [batch_size, image_width, image_height, 3]
        :param gt_labels: gt with classes. 4D tensor: [batch_size, image_width, image_height, 1]
        :param G_input_unsup: unsupervised input of the generator. 4D tensor: [batch_size, image_width, image_height, 3]
        :return:
                loss: final loss
                loss_fake: loss in recognising the data of the generator
                loss_real: loss on recognising the gt data
        """
        # Compute needed variables
        G_output_sup = self.G(G_input_sup)
        if G_input_unsup is None:
            D_input_fake = tf.nn.softmax(G_output_sup)
            target_mask_sup = target_mask = tf.squeeze(tf.less_equal(gt_labels, self.num_classes - 1), axis=3)
        else:
            G_output_unsup = self.G(G_input_unsup)
            D_input_fake = tf.concat([tf.nn.softmax(G_output_sup), tf.nn.softmax(G_output_unsup)], axis=0)
            shape = tf.shape(gt_labels)
            target_mask_unsup = tf.fill(value=True, dims=[shape[0], shape[1], shape[2]])
            target_mask_sup = tf.squeeze(tf.less_equal(gt_labels, self.num_classes - 1), axis=3)
            target_mask = tf.concat([target_mask_sup, target_mask_unsup], axis=0)

        D_input_real = tensor_utils.convert_label2onehot(gt_labels, self.num_classes)

        D_out_fake = self.D(D_input_fake)
        D_out_real = self.D(D_input_real)
        if self.use_lsgan:  # use mean squared error
            loss_real = tf.squared_difference(D_out_real, REAL_LABEL)
            loss_fake = tf.square(D_out_fake)
        else:  # use cross entropy
            # data coming from dataset. Discriminator should predict that it is real.
            loss_real = tf.nn.sigmoid_cross_entropy_with_logits(logits=D_out_real, labels=tf.ones_like(D_out_real))

            # data coming from generator. Discriminator should predict that it is fake->zeros
            loss_fake = tf.nn.sigmoid_cross_entropy_with_logits(logits=D_out_fake, labels=tf.zeros_like(D_out_fake))

        # Apply mask to loss to ignore not valid value that are present in the gt
        loss_real = tf.reduce_mean(tf.boolean_mask(loss_real, target_mask_sup)) / 2
        loss_fake = tf.reduce_mean(tf.boolean_mask(loss_fake, target_mask)) / 2

        # Compute final loss
        loss = loss_fake + loss_real

        # Summary
        summary = []
        if self.save_summary_images:
            D_out_fake_slice = tf.slice(D_out_fake, [0, 0, 0, 0], [1, self.image_height, self.image_width, 1])
            D_out_real_slice = tf.slice(D_out_real, [0, 0, 0, 0], [1, self.image_height, self.image_width, 1])
            gt_slice = image_utils.convert_gt2rgb(
                tf.slice(gt_labels, [0, 0, 0, 0], [1, self.image_height, self.image_width, 1]))
            image_slice = image_utils.convert_output2rgb(
                tf.slice(G_output_sup, [0, 0, 0, 0], [1, self.image_height, self.image_width, self.num_classes]))

            summary = [tf.summary.image('D/fake_in', image_slice),
                       tf.summary.image('D/real_in', gt_slice),
                       tf.summary.image('D/fake_out', D_out_fake_slice),
                       tf.summary.image('D/real_out', D_out_real_slice)]
        summary.append(tf.summary.scalar('loss_D/D', loss))
        summary.append(tf.summary.scalar('loss_D/D_fake', loss_fake))
        summary.append(tf.summary.scalar('loss_D/D_real', loss_real))
        return loss, summary

    def adversarial_loss(self, G_input, gt_labels=None):
        """
        Compute the adversarial loss provided by the discriminator net

        :param G_input: input of the generator. 4D tensor: [batch_size, image_width, image_height, 3]
        :param gt_labels: gt with classes. 4D tensor: [batch_size, image_width, image_height, 1]
        :return: adversarial loss provided by discriminator
        """
        # Compute needed variables
        G_output = self.G(G_input)
        D_out_fake = self.D(tf.nn.softmax(G_output))

        if self.use_lsgan:
            loss = tf.squared_difference(D_out_fake, REAL_LABEL)
        else:
            # data coming from generator. Generator wants to fool discriminator and let it predict that it is real->ones (although it is fake data)
            loss = tf.nn.sigmoid_cross_entropy_with_logits(logits=D_out_fake, labels=tf.ones_like(D_out_fake))

        # Unsupervised learning doesn't have gt
        if gt_labels is not None:
            target_mask = tf.squeeze(tf.less_equal(gt_labels, self.num_classes - 1), axis=3)
            loss = tf.boolean_mask(loss, target_mask)

        loss = tf.reduce_mean(loss)
        return loss

    def generator_loss_adv(self, G_input, gt_labels, lambda_adv):
        """
        Compute the loss of the generator with the addition of adversarial loss
        :param G_input: input of the generator. 4D tensor: [batch_size, image_width, image_height, 3]
        :param gt_labels: gt with classes. 4D tensor: [batch_size, image_width, image_height, 1]
        :param lambda_adv: measure how to weight the adversarial loss in the final loss
        :return:
                loss: final loss
                loss_ce: generator loss
                loss_adv: adversarial loss
        """

        # Compute the different losses
        loss_ce, sup_summary = self.generator_loss_supervised(G_input=G_input, gt_labels=gt_labels)
        loss_adv = lambda_adv * self.adversarial_loss(G_input=G_input, gt_labels=gt_labels)

        # Compose the final loss as a weighted sum of all the terms
        loss = loss_ce + loss_adv

        # Summary
        summary = sup_summary
        summary.append(tf.summary.scalar('loss_G/G_sup_adv', loss_adv))

        return loss, summary

    def generator_loss_unsupervised(self):
        """
        Compute the loss of the generator in an unsupervised manner

        :param G_input_unsup: input of the generator. 4D tensor: [batch_size, image_width, image_height, 3]
        :return: unsupervised loss of the generator
                loss: final loss
                loss_adv: adversarial loss
                loss_semi: unsupervised loss
        """
        # Compute needed variables
        G_output_unsup = self.G(self.unsup_input)

        # Semi supervised loss
        D_out_fake = self.D(tf.nn.softmax(G_output_unsup))
        D_out_sigmoid = tf.nn.sigmoid(D_out_fake)
        synth_gt_labels = tf.argmax(G_output_unsup, axis=3)
        synth_gt_labels = tf.cast(synth_gt_labels, tf.uint8)
        synth_gt_one_hot = tf.one_hot(synth_gt_labels, self.num_classes)
        unsup_weight = tf.to_float(tf.equal(synth_gt_labels, 0)) * self.class_weights[0] + tf.to_float(
            tf.equal(synth_gt_labels, 1)) * self.class_weights[1] + tf.to_float(tf.equal(synth_gt_labels, 2)) * \
                                                                    self.class_weights[2] + tf.to_float(
            tf.equal(synth_gt_labels, 3)) * self.class_weights[
                                                                                                3] + tf.to_float(
            tf.equal(synth_gt_labels, 4)) * self.class_weights[4] + tf.to_float(
            tf.equal(synth_gt_labels, 5)) * self.class_weights[5] + tf.to_float(tf.equal(synth_gt_labels, 6)) * \
                                                                    self.class_weights[6] + tf.to_float(
            tf.equal(synth_gt_labels, 7)) * self.class_weights[
                                                                                                7] + tf.to_float(
            tf.equal(synth_gt_labels, 8)) * self.class_weights[8] + tf.to_float(
            tf.equal(synth_gt_labels, 9)) * self.class_weights[9] + tf.to_float(tf.equal(synth_gt_labels, 10)) * \
                                                                    self.class_weights[10] + tf.to_float(
            tf.equal(synth_gt_labels, 11)) * self.class_weights[
                                                                                                 11] + tf.to_float(
            tf.equal(synth_gt_labels, 12)) * self.class_weights[12] + tf.to_float(
            tf.equal(synth_gt_labels, 13)) * self.class_weights[13] + tf.to_float(tf.equal(synth_gt_labels, 14)) * \
                                                                      self.class_weights[14] + tf.to_float(
            tf.equal(synth_gt_labels, 15)) * self.class_weights[
                                                                                                   15] + tf.to_float(
            tf.equal(synth_gt_labels, 16)) * self.class_weights[16] + tf.to_float(
            tf.equal(synth_gt_labels, 17)) * self.class_weights[17] + tf.to_float(tf.equal(synth_gt_labels, 18)) * \
                                                                      self.class_weights[18]
        loss_ce = tf.losses.softmax_cross_entropy(logits=G_output_unsup, onehot_labels=synth_gt_one_hot,
                                                  weights=tf.multiply(unsup_weight,tf.squeeze(D_out_fake,axis=3)), reduction=Reduction.NONE)
        loss_ce_masked = tf.boolean_mask(loss_ce, self.expanded_mask)
        loss_unsup = tf.reduce_mean(loss_ce_masked)
        loss_semi = self.lambda_semi * loss_unsup

        # Adversarial loss
        loss_adv = self.lambda_adv_unsup * self.adversarial_loss(G_input=self.unsup_input)
        loss_tot = loss_adv + loss_semi

        # Summary
        summary = []
        if self.save_summary_images:
            image_slice = image_utils.convert_output2rgb(tf.slice(G_output_unsup, [0, 0, 0, 0],
                                                                  [1, self.image_height, self.image_width,
                                                                   self.num_classes]))
            D_out_slice = tf.slice(D_out_sigmoid, [0, 0, 0, 0], [1, self.image_height, self.image_width, 1])

            mask = tf.cast(self.expanded_mask, tf.uint8)
            mask = tf.expand_dims(mask, axis=3)
            mask_slice = tf.slice(mask * 255, [0, 0, 0, 0], [1, self.image_height, self.image_width, 1])

            summary = [tf.summary.image('unsup/G_output', image_slice),
                       tf.summary.image('unsup/D_output', D_out_slice),
                       tf.summary.image('unsup/mask', mask_slice)]

        summary.append(tf.summary.scalar('loss_G/G_unsup_tot', loss_tot))
        summary.append(tf.summary.scalar('loss_G/G_unsup_ce', loss_semi))
        summary.append(tf.summary.scalar('loss_G/G_unsup_adv', loss_adv))
        return loss_adv, loss_semi, summary

    def compute_validation_loss(self, G_input, gt_labels):
        """
        Compute the validation loss

        :param G_input: input of the generator. 4D tensor: [batch_size, image_width, image_height, 3]
        :param gt_labels: gt with classes. 4D tensor: [batch_size, image_width, image_height, 1]
        :return: validation loss
        """
        # Reformat input values
        G_input = tf.expand_dims(G_input, axis=0)
        gt_labels = tf.expand_dims(gt_labels, axis=0)

        # Compute loss
        loss, _ = self.generator_loss_supervised(G_input=G_input, gt_labels=gt_labels)

        return loss

    def compute_evaluation_metrics(self, G_input, gt_labels):
        """
        Compute all the metrics for the evaluation

        :param G_input: input of the generator. 4D tensor: [batch_size, image_width, image_height, 3]
        :param gt_labels: gt with classes. 4D tensor: [batch_size, image_width, image_height, 1]
        :return:
                mIoU: mean intersection over union
                mIou_update_op: update operation for mIoU
                accuracy: pixel accuracy
                accuracy_update_op: update operation for pixel accuracy
        """
        # predictions
        raw_output = self.G(G_input)
        raw_output = tf.argmax(raw_output, axis=3)
        pred = tf.expand_dims(raw_output, axis=3)
        pred = tf.reshape(pred, [-1, ])
        gt = tf.reshape(gt_labels, [-1, ])

        # Ignoring all labels greater than or equal to n_classes.
        mask = tf.less_equal(gt, self.num_classes - 1)
        pred = tf.boolean_mask(pred, mask)
        gt = tf.boolean_mask(gt, mask)

        # mIoU
        mIoU, mIou_update_op = tf.metrics.mean_iou(predictions=pred, labels=gt, num_classes=self.num_classes)

        # confusion matrix
        confusion_matrix = tf.confusion_matrix(predictions=pred, labels=gt, num_classes=self.num_classes)

        # Pixel accuracy
        accuracy, accuracy_update_op = tf.metrics.accuracy(predictions=pred, labels=gt)

        return mIoU, mIou_update_op, accuracy, accuracy_update_op, confusion_matrix
