import logging
import os
from datetime import datetime

import numpy as np
import tensorflow as tf

from model import GAN
from utils import tensor_utils

FLAGS = tf.flags.FLAGS

# Utilized data
tf.flags.DEFINE_string('training_file', '../../data/tfrecords/GTA/train_div_2.tfrecords',
                       'tfrecords file for training')
tf.flags.DEFINE_string('validation_file', '../../data/tfrecords/Cityscapes/test.tfrecords',
                       'tfrecords file for validation')
tf.flags.DEFINE_string('unsupervised_file', '../../data/tfrecords/Cityscapes/train.tfrecords',
                       'tfrecords file for unsupervised learning')
tf.flags.DEFINE_string('pretrained_folder', '../../data/pretrained', 'path of the folder for pretraied encoder models')

# Data format
tf.flags.DEFINE_integer('batch_size', 1, 'batch size')
tf.flags.DEFINE_integer('validation_batch_size', 1, ' validation batch size default: 1449')
tf.flags.DEFINE_integer('image_height', 375, 'image size, default: 256')
tf.flags.DEFINE_integer('image_width', 750, 'image size, default: 256')
tf.flags.DEFINE_integer('num_classes', 19, 'number of output classes')
tf.flags.DEFINE_string('norm', 'batch', '[instance, batch] use instance norm or batch norm, default: instance')

# Training parameters
tf.flags.DEFINE_string('train_type', 'semi-supervised',
                       '[supervised, adversarial, semi-supervised, domain-adaptation] type of training to use')
tf.flags.DEFINE_integer('training_steps', 40000, 'number of training steps')
tf.flags.DEFINE_integer('logging_steps_interval', 1, 'interval of steps between each log')
tf.flags.DEFINE_integer('validation_steps_interval', 1, 'interval of steps between each validation')
tf.flags.DEFINE_integer('save_steps_interval', 10000, 'interval of steps between each log')
tf.flags.DEFINE_integer('semi_start', 0, 'number of steps before start computing unsupervised loss (L-semi)')
tf.flags.DEFINE_boolean('use_lsgan', False,
                        'define the type of loss of the discriminator (least square or cross_entropy)')
tf.flags.DEFINE_boolean('separate_D_training', False,
                        'specifies if the discriminator must be trained in a separate time for real and fake data')
tf.flags.DEFINE_boolean('save_summary_images', False,
                        'specifies if the training images must be saved during the training (NB: higher memory usage)')

# Checkpoint folder
tf.flags.DEFINE_string('load_model',
                       None,
                       'folder of saved model that you wish to continue training (checkpoint/{})')

# Network optimization parameters
tf.flags.DEFINE_float('lambda_adv_sup', 0.01, 'weight for discriminator in supervised mode')
tf.flags.DEFINE_float('lambda_adv_unsup', 0.001, 'weight for discriminator in unsupervised mode')
tf.flags.DEFINE_float('lambda_unsup', 0.1, 'weight for unsupervised loss')
tf.flags.DEFINE_float('threshold_unsup', 0.2, 'threshold applied to confidence map produced by discriminator')

# Generator optimization parameter (generator uses SGD optimizer)
tf.flags.DEFINE_float('generator_learning_rate', 2.5e-4, 'initial learning rate for generator')
tf.flags.DEFINE_float('generator_end_learning_rate', 0.0, 'final decay rate for generator optimizer')
tf.flags.DEFINE_integer('generator_start_decay_step', 0, 'number of steps after which start decay for generator')
tf.flags.DEFINE_integer('generator_decay_steps', 40000, 'number of steps of decay learning rate for generator')
tf.flags.DEFINE_float('generator_decay_power', 0.9, 'power of weight decay for generator optimizer')
tf.flags.DEFINE_float('generator_momentum', 0.9, 'momentum for generator optimizer')
tf.flags.DEFINE_float('generator_weight_decay', 1e-4, 'weight decay for momentum of generator optimizer')

# Discriminator optimization parameter (discriminator uses Adam optimizer)
tf.flags.DEFINE_float('discriminator_learning_rate', 1e-4, 'initial learning rate for discriminator')
tf.flags.DEFINE_float('discriminator_end_learning_rate', 0.0, 'final decay rate for discriminator optimizer')
tf.flags.DEFINE_integer('discriminator_start_decay_step', 0,
                        'number of steps after which start decay for discriminator')
tf.flags.DEFINE_integer('discriminator_decay_steps', 40000, 'number of steps of decay learning rate for discriminator')
tf.flags.DEFINE_float('discriminator_decay_power', 0.9, 'power of weight decay for discriminator optimizer')
tf.flags.DEFINE_float('discriminator_momentum', 0.9, 'momentum for generator optimizer (utilized as beta1 parameter)')
tf.flags.DEFINE_float('discriminator_weight_decay', 1e-4, 'weight decay for discriminator optimizer')
tf.flags.DEFINE_float('threshold2', 0.9999, 'threshold for region growing')
k = 0
expanded_mask_slice = None
visited = None
mask_list = []


def expandMask(G_input, G_output, G_output_RGB, mask, threshold):
    global k
    global expanded_mask_slice
    expanded_mask_final = np.zeros([FLAGS.batch_size, FLAGS.image_height, FLAGS.image_width])
    for i in range(FLAGS.batch_size):
        expanded_mask_slice = np.full([FLAGS.image_height, FLAGS.image_width], False)
        G_output_slice = G_output[i, :, :, :]
        G_input_slice = G_input[i, :, :, :]
        G_output_RGB_slice = G_output_RGB[i, :, :, :]
        mask_slice = mask[i, :, :]
        labels = np.argmax(G_output_slice, axis=2)
        expandSingleMask(G_output_slice, labels, mask_slice, threshold)
        # scipy.misc.imsave('output_images/' + str(threshold) + '_' + str(k) + '_input.jpg', G_input_slice)
        # scipy.misc.imsave('output_images/' + str(threshold) + '_' + str(k) + '_labels.jpg', G_output_RGB_slice)
        # scipy.misc.imsave('output_images/' + str(threshold) + '_' + str(k) + '_mask.jpg', mask_slice.astype(int) * 255)
        # scipy.misc.imsave('output_images/' + str(threshold) + '_' + str(k) + '_mask_expanded.jpg',expanded_mask_slice.astype(int) * 255)
        k = k + 1
        expanded_mask_final[i, :, :] = expanded_mask_slice
    return expanded_mask_final


def expandSingleMask(G_out, labels, mask, threshold):
    global visited
    global mask_list
    visited = np.full([FLAGS.image_height, FLAGS.image_width], False)

    # fill the initial list
    for i in range(FLAGS.image_height):
        for j in range(FLAGS.image_width):
            if mask[i, j]:
                mask_list.append((i, j))
    # expand starting from all the white pixels
    while len(mask_list) > 0:
        current = mask_list.pop(0)
        i = current[0]
        j = current[1]

        if visited[i, j] == False:
            current_class = labels[i, j]
            visited[i, j] = True
            expanded_mask_slice[i, j] = True
            if j < FLAGS.image_width - 1 and labels[i, j + 1] == current_class and G_out[
                i, j + 1, current_class] > threshold and \
                    mask[i, j + 1] == False:
                mask_list.append((i, j + 1))
            if j > 0 and labels[i, j - 1] == current_class and G_out[i, j - 1, current_class] > threshold and mask[
                i, j - 1] == False:
                mask_list.append((i, j - 1))
            if i < FLAGS.image_height - 1 and labels[i + 1, j] == current_class and G_out[
                i + 1, j, current_class] > threshold and \
                    mask[i + 1, j] == False:
                mask_list.append((i + 1, j))
            if i > 0 and labels[i - 1, j] == current_class and G_out[i - 1, j, current_class] > threshold and mask[
                i - 1, j] == False:
                mask_list.append((i - 1, j))


def train():
    # define mapping vector for domain adaptation
    #             [0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20]
    # class_mask = [0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0]
    class_mask = None


    class_weights_GTA_Val = [0.637960276015760, 0.906876401511850, 0.808888109204147, 0.979417971165853,
                             0.992849926507391, 0.988065518323295, 0.998489616496333, 0.999064071693913,
                             0.914391166506145, 0.975514686906170, 0.848631300794846, 0.995948128310475,
                             0.999638046230835, 0.971915722873151, 0.987760747283702, 0.995798327177682,
                             0.999222562735153, 0.999635347188153, 0.999932073075146]

    # Prevent errors
    assert FLAGS.train_type == 'supervised' or FLAGS.train_type == 'adversarial' or FLAGS.train_type == 'semi-supervised' or FLAGS.train_type == 'domain-adaptation', 'value not recognised! Allowed value are: [supervised, adversarial, semi-supervised]'
    if FLAGS.train_type == 'domain-adaptation':
        assert FLAGS.load_model is not None, 'Needed pre-trained model for domain adaptation!'

    # Define checkpoint folder name
    if FLAGS.load_model is not None:
        checkpoints_dir = "checkpoints/" + FLAGS.load_model.lstrip("checkpoints/")
    else:
        current_time = datetime.now().strftime("%Y%m%d-%H%M") \
                       + '_semi-' + str(FLAGS.semi_start) \
                       + '_G-lr-' + str(FLAGS.generator_learning_rate) \
                       + '_D-lr-' + str(FLAGS.discriminator_learning_rate) \
                       + '_yas-' + str(FLAGS.lambda_adv_sup) \
                       + '_yau-' + str(FLAGS.lambda_adv_unsup) \
                       + '_yu-' + str(FLAGS.lambda_unsup) \
                       + '_t-' + str(FLAGS.threshold_unsup)
        checkpoints_dir = "checkpoints/{}".format(current_time)
        try:
            os.makedirs(checkpoints_dir)
        except os.error:
            pass

    # Define the model
    graph = tf.Graph()
    with graph.as_default():
        tf.set_random_seed(12345)
        gan = GAN(
            train_file=FLAGS.training_file,
            validation_file=FLAGS.validation_file,
            unsupervised_file=FLAGS.unsupervised_file,
            batch_size=FLAGS.batch_size,
            image_height=FLAGS.image_height,
            image_width=FLAGS.image_width,
            num_classes=FLAGS.num_classes,
            pretrained_folder=FLAGS.pretrained_folder,
            lambda_adv_sup=FLAGS.lambda_adv_sup,
            lambda_adv_unsup=FLAGS.lambda_adv_unsup,
            lambda_unsup=FLAGS.lambda_unsup,
            threshold_unsup=FLAGS.threshold_unsup,
            generator_learning_rate=FLAGS.generator_learning_rate,
            generator_momentum=FLAGS.generator_momentum,
            generator_weight_decay=FLAGS.generator_weight_decay,
            generator_decay_power=FLAGS.generator_decay_power,
            generator_decay_steps=FLAGS.generator_decay_steps,
            generator_start_decay_step=FLAGS.generator_start_decay_step,
            generator_end_learning_rate=FLAGS.generator_end_learning_rate,
            discriminator_learning_rate=FLAGS.discriminator_learning_rate,
            discriminator_momentum=FLAGS.discriminator_momentum,
            discriminator_weight_decay=FLAGS.discriminator_weight_decay,
            discriminator_decay_power=FLAGS.discriminator_decay_power,
            discriminator_decay_steps=FLAGS.discriminator_decay_steps,
            discriminator_start_decay_step=FLAGS.discriminator_start_decay_step,
            discriminator_end_learning_rate=FLAGS.discriminator_end_learning_rate,
            train_type=FLAGS.train_type,
            use_lsgan=FLAGS.use_lsgan,
            save_summary_images=FLAGS.save_summary_images,
            class_weights=class_weights_GTA_Val
        )

        G_sup_loss, G_composed_loss_adv, G_composed_loss_adv_semi, D_loss, G_optimizer_sup, G_optimizer_unsup_adv_semi, D_optimizer, training_summary_sup, training_summary_unsup, G_in_unsup, G_output_unsup, G_output_unsup_RGB, threshold_mask = gan.training_model()
        iterator, mean_validation_loss, mean_validation_update_op, confusion_matrix, accuracy, accuracy_update_op, mean_IoU, mean_IoU_update_op, validation_summary = gan.validation_model()

        train_writer = tf.summary.FileWriter(checkpoints_dir, graph)
        saver = tf.train.Saver()
    summary = None
    G_loss_val = 0
    D_loss_val = 0
    step = 0
    start_step = 0
    max_mIoU = 0

    logging.info(' training type: {}'.format(FLAGS.train_type))
    logging.info(' least square: {}'.format(FLAGS.use_lsgan))
    logging.info(' discriminator learning rate: {}'.format(FLAGS.discriminator_learning_rate))
    logging.info(' lambda adv sup: {}'.format(FLAGS.lambda_adv_sup))
    logging.info(' lambda adv unsup: {}'.format(FLAGS.lambda_adv_unsup))
    logging.info(' lambda unsup: {}'.format(FLAGS.lambda_unsup))

    # Load checkpoint if provided
    with tf.Session(graph=graph, config=None) as sess:
        tf.set_random_seed(12345)
        if FLAGS.load_model is not None:  # Restore pre trained checkpoint
            sess.run(tf.local_variables_initializer())
            sess.run(tf.global_variables_initializer())
            checkpoint = tf.train.get_checkpoint_state(checkpoints_dir)
            meta_graph_path = checkpoint.model_checkpoint_path + ".meta"
            if FLAGS.train_type == 'domain-adaptation':  # load only the needed components of the networks
                restore_variables = gan.G.all_var + gan.D.variables
                loader = tf.train.Saver(var_list=restore_variables, reshape=True)
                loader.restore(sess, tf.train.latest_checkpoint(checkpoints_dir))
            else:
                loader = tf.train.import_meta_graph(meta_graph_path)
                loader.restore(sess, tf.train.latest_checkpoint(checkpoints_dir))
            start_step = int(meta_graph_path.split("-")[-1].split(".")[0])
            step = start_step
        else:  # Load pre trained encoder
            sess.run(tf.global_variables_initializer())
            sess.run(tf.local_variables_initializer())
            loader = tf.train.Saver(var_list=gan.G.encoder_var, reshape=True)
            loader.restore(sess, FLAGS.pretrained_folder + '/deeplab_resnet_init.ckpt')

        coord = tf.train.Coordinator()
        threads = tf.train.start_queue_runners(sess=sess, coord=coord)

        # Train
        try:
            while not coord.should_stop() and step < FLAGS.training_steps:
                # Separate different types of training
                if FLAGS.train_type == 'supervised':
                    _, G_loss_val, summary = (
                        sess.run(
                            [G_optimizer_sup, G_sup_loss, training_summary_sup]
                        )
                    )
                elif FLAGS.train_type == 'adversarial':
                    # Train generator
                    _, G_loss_val, summary = (
                        sess.run(
                            [G_optimizer_sup, G_sup_loss, training_summary_sup]
                        )
                    )
                    train_writer.add_summary(summary, step)
                    train_writer.flush()
                    # Train discriminator
                    _, D_loss_val, summary = (
                        sess.run(
                            [D_optimizer, D_loss, training_summary_sup]
                        )
                    )
                elif FLAGS.train_type == 'semi-supervised':
                    # Do semi first (As in the paper)
                    if step < FLAGS.semi_start:
                        # Compute the expanded mask
                        G_input, G_output, G_output_RGB, mask = (
                            sess.run(
                                [G_in_unsup, G_output_unsup, G_output_unsup_RGB, threshold_mask],
                                feed_dict={gan.lambda_semi: FLAGS.lambda_unsup}
                            )
                        )
                        expanded_mask = expandMask(G_input, G_output, G_output_RGB, mask, FLAGS.threshold2)
                        # Train only with unsupervised adversarial loss
                        _, G_loss_val, summary_sup, summary_unsup = (
                            sess.run(
                                [G_optimizer_unsup_adv_semi, G_composed_loss_adv_semi, training_summary_sup,
                                 training_summary_unsup], feed_dict={gan.lambda_semi: 0.0, gan.unsup_input: G_input,
                                                                     gan.expanded_mask: expanded_mask, }
                            )
                        )
                    else:
                        # Compute the expanded mask
                        G_input, G_output, G_output_RGB, mask = (
                            sess.run(
                                [G_in_unsup, G_output_unsup, G_output_unsup_RGB, threshold_mask],
                                feed_dict={gan.lambda_semi: FLAGS.lambda_unsup}
                            )
                        )

                        expanded_mask = expandMask(G_input, G_output, G_output_RGB, mask, FLAGS.threshold2)

                        # Train with unsupervised adversarial loss and unsupervised cross-entropy (L-semi)
                        _, G_loss_val, summary_sup, summary_unsup = (
                            sess.run(
                                [G_optimizer_unsup_adv_semi, G_composed_loss_adv_semi, training_summary_sup,
                                 training_summary_unsup],
                                feed_dict={gan.lambda_semi: FLAGS.lambda_unsup, gan.expanded_mask: expanded_mask,
                                           gan.unsup_input: G_input}
                            )
                        )
                    train_writer.add_summary(summary_sup, step)
                    train_writer.add_summary(summary_unsup, step)
                    train_writer.flush()
                    _, D_loss_val, summary = (
                        sess.run(
                            [D_optimizer, D_loss, training_summary_sup]
                        )
                    )

                train_writer.add_summary(summary, step)
                train_writer.flush()

                # Output computed metrics
                if step % FLAGS.logging_steps_interval == 0:
                    logging.info('-----------Step %d:-------------' % step)
                    if FLAGS.train_type == 'supervised':
                        logging.info(' G_loss : {}'.format(G_loss_val))

                    elif FLAGS.train_type == 'adversarial' or FLAGS.train_type == 'semi-supervised' or FLAGS.train_type == 'domain-adaptation':
                        logging.info(' G_loss : {}'.format(G_loss_val))
                        logging.info(' D_loss : {}'.format(D_loss_val))

                # Compute validation metrics
                if step % FLAGS.validation_steps_interval == 0:
                    sess.run(tf.local_variables_initializer())
                    sess.run([iterator.initializer])
                    n = 0
                    local_confusion_matrix = np.zeros((FLAGS.num_classes, FLAGS.num_classes), dtype=np.int)
                    for i in range(FLAGS.validation_batch_size):
                        try:
                            cm, _, _, _ = sess.run(
                                [confusion_matrix, mean_validation_update_op, mean_IoU_update_op, accuracy_update_op])
                            local_confusion_matrix += cm
                            n += 1
                        except tf.errors.OutOfRangeError:  # no more items
                            logging.info(
                                ' The chosen value ({}) exceed number of validation images! Used {} instead.'.format(
                                    FLAGS.validation_batch_size, n))
                            break
                    mean_validation_loss_val, mean_IoU_val, accuracy_val, summary = sess.run(
                        [mean_validation_loss, mean_IoU, accuracy, validation_summary])
                    train_writer.add_summary(summary, step)
                    train_writer.flush()
                    logging.info(' Validation_loss : {}'.format(mean_validation_loss_val))
                    logging.info(' accuracy : {:0.2f}%'.format(accuracy_val * 100))
                    tensor_utils.compute_and_print_IoU_per_class(local_confusion_matrix, FLAGS.num_classes,
                                                                 class_mask=class_mask)
                    # Save the weight of the network
                    if step % FLAGS.save_steps_interval == 0 and mean_IoU_val > max_mIoU:
                        save_path = saver.save(sess, checkpoints_dir + "/model.ckpt", global_step=step)
                        logging.info("Model saved in file: %s" % save_path)
                        max_mIoU = mean_IoU_val

                step += 1

        except KeyboardInterrupt:
            logging.info('Interrupted')
            coord.request_stop()
        except Exception as e:
            coord.request_stop(e)
        finally:
            # Output computed metrics
            logging.info('-----------Step %d:-------------' % step)
            if FLAGS.train_type == 'supervised':
                logging.info(' G_loss : {}'.format(G_loss_val))

            elif FLAGS.train_type == 'adversarial' or FLAGS.train_type == 'semi-supervised' or FLAGS.train_type == 'domain-adaptation':
                logging.info(' G_loss : {}'.format(G_loss_val))
                logging.info(' D_loss : {}'.format(D_loss_val))

            # Compute validation metrics
            sess.run(tf.local_variables_initializer())
            sess.run([iterator.initializer])
            n = 0
            local_confusion_matrix = np.zeros((FLAGS.num_classes, FLAGS.num_classes), dtype=np.int)
            for i in range(FLAGS.validation_batch_size):
                try:
                    cm, _, _, _ = sess.run(
                        [confusion_matrix, mean_validation_update_op, mean_IoU_update_op, accuracy_update_op])
                    local_confusion_matrix += cm
                    n += 1
                except tf.errors.OutOfRangeError:  # no more items
                    logging.info(
                        ' The chosen value ({}) exceed number of validation images! Used {} instead.'.format(
                            FLAGS.validation_batch_size, n))
                    break
            mean_validation_loss_val, mean_IoU_val, accuracy_val, summary = sess.run(
                [mean_validation_loss, mean_IoU, accuracy, validation_summary])
            train_writer.add_summary(summary, step)
            train_writer.flush()
            logging.info(' Validation_loss : {}'.format(mean_validation_loss_val))
            logging.info(' accuracy : {:0.2f}%'.format(accuracy_val * 100))
            tensor_utils.compute_and_print_IoU_per_class(local_confusion_matrix, FLAGS.num_classes,
                                                         class_mask=class_mask)
            if FLAGS.train_type != 'domain-adaptation':
                save_path = saver.save(sess, checkpoints_dir + "/model.ckpt", global_step=step)
                logging.info("Model saved in file: %s" % save_path)

            # When done, ask the threads to stop.
            coord.request_stop()
            coord.join(threads)


def main(unused_argv):
    train()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    tf.app.run()
