import tensorflow as tf

from network import Deeplab_v2
from utils import image_utils


class Generator:
    def __init__(self, name, num_classes, image_height, image_width, pretrained_folder, weight_decay):
        self.name = name
        self.reuse = tf.AUTO_REUSE
        self.num_classes = num_classes
        self.image_height = image_height
        self.image_width = image_width
        self.pretrained_folder = pretrained_folder
        self.first_run = True
        self.weight_decay = weight_decay

    def __call__(self, net_input):
        """
        Base method for calling the generator
        :param net_input: input of the network. 4D tensor: [batch_size, image_width, image_height, 3]
        :return: output of the network. 4D tensor: [batch_size, image_width, image_height, num_classes]
        """

        with tf.name_scope(self.name):
            tf.set_random_seed(12345)
            net = Deeplab_v2(net_input, self.num_classes, True, self.first_run).outputs
            # Trainable Variables
            all_trainable = tf.global_variables()

            # Fine-tune part
            encoder_trainable = [v for v in all_trainable if 'fc' not in v.name]  # lr * 1.0
            # Decoder part
            decoder_trainable = [v for v in all_trainable if 'fc' in v.name]

            self.first_run = False

        encoder_var_filtered = []
        for var in encoder_trainable:
            if "Variable" not in var.name and "optimizer" not in var.name and "D" not in var.name and "power" not in var.name:
                encoder_var_filtered.append(var)

        all_var_filtered = []
        for var in all_trainable:
            if "Variable" not in var.name and "optimizer" not in var.name and "D" not in var.name and "power" not in var.name:
                all_var_filtered.append(var)

        # L2 regularization
        l2_losses = [self.weight_decay * tf.nn.l2_loss(v) for v in all_var_filtered if 'weights' in v.name]
        self.l2_loss = tf.add_n(l2_losses)

        self.encoder_var = encoder_var_filtered
        self.all_var = all_var_filtered

        return net

    def sample(self, net_input):
        """
        Method called to freeze graph
        :param net_input: input of the generator network
        :return: output image produced by the network
        """
        image = self.__call__(net_input)
        image = image_utils.convert_output2rgb(image)
        image = tf.image.encode_png(tf.squeeze(image, [0]))
        return image
